from torch.utils.tensorboard import SummaryWriter

from tez.callbacks import Callback


class TensorBoardLogger(Callback):
    def __init__(self, log_dir=".logs/"):
        self.writer = SummaryWriter(log_dir=log_dir, flush_secs=30)

    def on_train_epoch_end(self, model):
        for metric in model.metrics["train"]:
            self.writer.add_scalar(f"train/{metric}", model.metrics["train"][metric], model.current_epoch)

    def on_valid_epoch_end(self, model):
        for metric in model.metrics["valid"]:
            self.writer.add_scalar(f"valid/{metric}", model.metrics["valid"][metric], model.current_epoch)



import sys


try:
    from loguru import logger

    logger.configure(handlers=[dict(sink=sys.stderr, format="> <level>{level:<7} {message}</level>")])
except ImportError:

    import logging

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))
    logger = logging.getLogger("tez")
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)



from tqdm.auto import tqdm

from tez.callbacks import Callback


class Progress(Callback):
    def __init__(self, num_train_steps, num_valid_steps):
        self.num_train_steps = num_train_steps
        self.num_valid_steps = num_valid_steps
        self._train_tqdm = None
        self._valid_tqdm = None

    def on_train_start(self, tez_trainer, **kwargs):
        self._train_tqdm = tqdm(total=self.num_train_steps)
        # if self.num_valid_steps:
        #    self._valid_tqdm = tqdm(total=self.num_valid_steps, leave=False)

    def format_metrics(self, metrics, stage):
        metrics_str = ", ".join(["{}={:.4f}".format(k, v) for k, v in metrics.items() if k not in ("epoch", "steps")])
        if stage == "train":
            return f"[{stage}] {metrics_str}"
        elif stage == "valid":
            return f"[{stage}] {metrics_str}"
        else:
            raise ValueError(f"Unknown stage: {stage}")

    def on_train_step_end(self, tez_trainer, **kwargs):
        train_metrics = tez_trainer.metrics["train"]
        if "epoch" in train_metrics:
            del train_metrics["epoch"]
        self._train_tqdm.set_postfix(epoch=tez_trainer.current_epoch, **train_metrics)
        self._train_tqdm.update(1)

    # def on_valid_step_end(self, tez_trainer, **kwargs):
    #     if self.num_valid_steps is None:
    #         return
    #     self._valid_tqdm.update(1)

    def on_valid_epoch_end(self, tez_trainer, **kwargs):
        # check if there are any validation metrics
        if len(tez_trainer.metrics["valid"]) == 0:
            return

        train_metrics = tez_trainer.metrics["train"]
        epoch = tez_trainer.current_epoch
        steps = tez_trainer.current_train_step
        # train_metrics["batches"] = tez_trainer.current_train_step
        train_metrics = self.format_metrics(train_metrics, stage="train")
        # self._train_tqdm.write(train_metrics)

        valid_metrics = tez_trainer.metrics["valid"]
        valid_metrics["epoch"] = tez_trainer.current_epoch
        # valid_metrics["batches"] = tez_trainer.current_train_step
        valid_metrics = self.format_metrics(valid_metrics, stage="valid")

        metrics_string = f"{train_metrics} {valid_metrics} [e={epoch} steps={steps}]"
        self._train_tqdm.write(metrics_string)

    def on_train_end(self, tez_trainer, **kwargs):
        self._train_tqdm.close()



from typing import List


class Callback:
    def on_epoch_start(self, tez_trainer, **kwargs):
        return

    def on_epoch_end(self, tez_trainer, **kwargs):
        return

    def on_train_epoch_start(self, tez_trainer, **kwargs):
        return

    def on_train_epoch_end(self, tez_trainer, **kwargs):
        return

    def on_valid_epoch_start(self, tez_trainer, **kwargs):
        return

    def on_valid_epoch_end(self, tez_trainer, **kwargs):
        return

    def on_train_step_start(self, tez_trainer, **kwargs):
        return

    def on_train_step_end(self, tez_trainer, **kwargs):
        return

    def on_valid_step_start(self, tez_trainer, **kwargs):
        return

    def on_valid_step_end(self, tez_trainer, **kwargs):
        return

    def on_test_step_start(self, tez_trainer, **kwargs):
        return

    def on_test_step_end(self, tez_trainer, **kwargs):
        return

    def on_train_start(self, tez_trainer, **kwargs):
        return

    def on_train_end(self, tez_trainer, **kwargs):
        return


class CallbackRunner:
    def __init__(self, callbacks: List[Callback], tez_trainer):
        self.tez_trainer = tez_trainer
        self.callbacks = callbacks

    def __call__(self, current_state, **kwargs):
        for cb in self.callbacks:
            _ = getattr(cb, current_state.value)(self.tez_trainer, **kwargs)



import multiprocessing
import os
import time
import warnings

import torch
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler

from tez import enums
from tez.callbacks import CallbackRunner, Progress
from tez.logger import logger
from tez.utils import AverageMeter

from .config import TezConfig


warnings.filterwarnings("ignore", category=UserWarning)


class Tez:
    def __init__(self, model):
        self.model = model
        self.config = None
        self.train_dataset = None
        self.valid_dataset = None
        self.optimizer = None
        self.scheduler = None
        self.scaler = None
        self.current_epoch = 0
        self.batch_index = 0
        self.current_train_step = 0
        self.current_valid_step = 0
        self.num_gpu = 0
        self.local_rank = -1
        self.world_size = 1
        self._model_state = None
        self._train_state = None
        self.metrics_meter = None
        self.metrics = {}
        self.metrics["train"] = {}
        self.metrics["valid"] = {}
        self.metrics["test"] = {}
        self.num_train_steps = None
        self.num_valid_steps = None

    def _configure_model(self):
        local_rank = int(os.environ.get("LOCAL_RANK", -1))
        if local_rank != -1 and local_rank != self.local_rank:
            self.local_rank = local_rank

        if self.config.device == "cpu":
            device = torch.device("cpu")
            self.num_gpu = 0
        elif self.config.device == "cuda":
            if torch.cuda.device_count() > 1:
                if self.local_rank == -1:
                    device = torch.device("cuda:0")
                    self.num_gpu = torch.cuda.device_count()
                    self.model = torch.nn.DataParallel(self.model, device_ids=list(range(self.num_gpu)))
                else:
                    torch.distributed.init_process_group(backend="nccl")
                    self.world_size = torch.distributed.get_world_size()
                    device = torch.device("cuda", self.local_rank)
                    self.model.to(device)
                    self.model = torch.nn.parallel.DistributedDataParallel(
                        self.model,
                        device_ids=[self.local_rank],
                        output_device=self.local_rank,
                    )
                    self.num_gpu = 1
            else:
                logger.info("Using single GPU")
                device = torch.device("cuda:0")
                self.num_gpu = 1

        if self.local_rank == -1:
            self.model.to(device)

    def _init_trainer(self, train_dataset, valid_dataset, config, **kwargs):
        self.config = config
        self.train_dataset = train_dataset
        self.valid_dataset = valid_dataset

        if "train_loader" in kwargs:
            self.train_loader = kwargs["train_loader"]
        else:
            self.train_loader = None
        if "valid_loader" in kwargs:
            self.valid_loader = kwargs["valid_loader"]
        else:
            self.valid_loader = None

        if "train_sampler" in kwargs:
            self.train_sampler = kwargs["train_sampler"]
        else:
            self.train_sampler = None

        if "valid_sampler" in kwargs:
            self.valid_sampler = kwargs["valid_sampler"]
        else:
            self.valid_sampler = None

        if "train_collate_fn" in kwargs:
            self.train_collate_fn = kwargs["train_collate_fn"]
        else:
            self.train_collate_fn = None

        if "valid_collate_fn" in kwargs:
            self.valid_collate_fn = kwargs["valid_collate_fn"]
        else:
            self.valid_collate_fn = None

        self.num_train_steps = int(len(self.train_dataset) / self.config.training_batch_size * self.config.epochs)
        if self.valid_dataset:
            self.num_valid_steps = int(len(self.valid_dataset) / self.config.validation_batch_size)
        else:
            self.num_valid_steps = None

        _progress = Progress(num_train_steps=self.num_train_steps, num_valid_steps=self.num_valid_steps)

        if "callbacks" in kwargs:
            self.callbacks = [_progress] + kwargs["callbacks"]
        else:
            self.callbacks = [_progress]

        if self.config.num_jobs == -1:
            self.config.num_jobs = multiprocessing.cpu_count()
            if self.config.num_jobs > 4:
                self.config.num_jobs -= 2

        if self.world_size > 1 and self.train_sampler is None:
            self.train_sampler = DistributedSampler(
                self.train_dataset,
                num_replicas=self.world_size,
                rank=self.local_rank,
            )

        if self.world_size > 1 and self.valid_sampler is None:
            self.valid_sampler = DistributedSampler(
                self.valid_dataset,
                num_replicas=self.world_size,
                rank=self.local_rank,
            )

        if self.train_loader is None:
            self.train_loader = DataLoader(
                self.train_dataset,
                batch_size=self.config.training_batch_size,
                num_workers=self.config.num_jobs,
                sampler=self.train_sampler,
                shuffle=self.config.train_shuffle,
                collate_fn=self.train_collate_fn,
                drop_last=self.config.train_drop_last,
                pin_memory=self.config.pin_memory,
            )

        if self.valid_loader is None:
            if self.valid_dataset is not None:
                self.valid_loader = DataLoader(
                    self.valid_dataset,
                    batch_size=self.config.validation_batch_size,
                    num_workers=self.config.num_jobs,
                    sampler=self.valid_sampler,
                    shuffle=self.config.valid_shuffle,
                    collate_fn=self.valid_collate_fn,
                    drop_last=self.config.valid_drop_last,
                    pin_memory=self.config.pin_memory,
                )

        self.optimizer, self.scheduler = self.model.optimizer_scheduler()

        if self.config.fp16:
            self.scaler = torch.cuda.amp.GradScaler()

        self._callback_runner = CallbackRunner(self.callbacks, self)
        self._configure_model()
        self.train_state = enums.TrainingState.TRAIN_START

        if self.optimizer is None:
            raise Exception("No optimizer found")

        if self.local_rank != -1:
            if torch.distributed.get_rank() == 0:
                logger.info(f"\n{self.config}")
                if self.scheduler is None:
                    logger.warning("No scheduler found. Continuing without scheduler")
        else:
            logger.info(f"\n{self.config}")
            if self.scheduler is None:
                logger.warning("No scheduler found. Continuing without scheduler")

    def _init_load_weights(self, config):
        self.config = config
        if self.config.device == "cpu":
            device = torch.device("cpu")
        elif self.config.device == "cuda":
            device = torch.device("cuda:0")
        else:
            raise Exception("Unknown device. Please use 'cpu' or 'cuda'")

        if next(self.model.parameters()).device != device:
            self.model.to(device)

        return device

    @property
    def model_state(self):
        return self._model_state

    @model_state.setter
    def model_state(self, value):
        self._model_state = value

    @property
    def train_state(self):
        return self._train_state

    @train_state.setter
    def train_state(self, value):
        self._train_state = value
        if self._callback_runner is not None:
            if self.local_rank != -1:
                if torch.distributed.get_rank() == 0:
                    self._callback_runner(value)
            else:
                self._callback_runner(value)

    def name_to_metric(self, metric_name):
        if metric_name == "current_epoch":
            return self.current_epoch
        v_1 = metric_name.split("_")[0]
        v_2 = "_".join(metric_name.split("_")[1:])
        return self.metrics[v_1][v_2]

    def update_metrics(self, losses, monitor):
        if self._model_state.value == "end":
            return
        self.metrics[self._model_state.value].update(monitor)
        self.metrics[self._model_state.value]["loss"] = losses.avg

    def save(self, model_path, weights_only=False):
        if self.local_rank != -1 or self.num_gpu > 1:
            model_state_dict = self.model.module.state_dict()
        else:
            model_state_dict = self.model.state_dict()

        if weights_only:
            if self.local_rank != -1 and self.num_gpu > 1:
                if torch.distributed.get_rank() == 0:
                    torch.save(model_state_dict, model_path)
            else:
                torch.save(model_state_dict, model_path)
            return

        if self.optimizer is not None:
            opt_state_dict = self.optimizer.state_dict()
        else:
            opt_state_dict = None

        if self.scheduler is not None:
            sch_state_dict = self.scheduler.state_dict()
        else:
            sch_state_dict = None

        model_dict = {}
        model_dict["state_dict"] = model_state_dict
        model_dict["optimizer"] = opt_state_dict
        model_dict["scheduler"] = sch_state_dict
        model_dict["config"] = self.config

        if self.local_rank != -1 and self.num_gpu > 1:
            if torch.distributed.get_rank() == 0:
                torch.save(model_dict, model_path)
        else:
            torch.save(model_dict, model_path)

    def load(self, model_path, weights_only=False, config: TezConfig = None):
        if config is None:
            config = TezConfig()

        device = self._init_load_weights(config)

        model_dict = torch.load(model_path, map_location=device)
        if weights_only:
            self.model.load_state_dict(model_dict)
        else:
            self.model.load_state_dict(model_dict["state_dict"])

    def model_fn(self, data):
        for key, value in data.items():
            data[key] = value.to(self.config.device)
        if self.config.fp16:
            with torch.cuda.amp.autocast():
                output, loss, metrics = self.model(**data)
        else:
            output, loss, metrics = self.model(**data)
        return output, loss, metrics

    def _zero_grad(self):
        if self.config.gradient_accumulation_steps == 1 and self.batch_index == 0:
            self.model.zero_grad()

    def _backward(self, loss, metrics):
        if self.num_gpu > 1:
            loss = loss.mean()
            for metric in metrics:
                metrics[metric] = metrics[metric].mean()
        loss = loss / self.config.gradient_accumulation_steps

        if self.config.fp16:
            self.scaler.scale(loss).backward()
        else:
            loss.backward()

    def _clip_grad_norm(self):
        if self.config.clip_grad_norm != -1:
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.config.clip_grad_norm)

    def _step(self):
        is_bi_mod_acc_zero = (self.batch_index + 1) % self.config.gradient_accumulation_steps == 0
        is_bi_end = self.batch_index + 1 == self.train_loader
        if is_bi_mod_acc_zero or is_bi_end:
            if self.config.fp16:
                self.scaler.step(self.optimizer)
                self.scaler.update()
            else:
                self.optimizer.step()

            if self.scheduler is not None:
                if self.config.step_scheduler_after == "batch":
                    if self.config.step_scheduler_metric is None:
                        self.scheduler.step()
                    else:
                        step_metric = self.name_to_metric(self.config.step_scheduler_metric)
                        self.scheduler.step(step_metric)

            self.model.zero_grad()

    def train_step(self, data):
        self._zero_grad()
        _, loss, metrics = self.model_fn(data)
        self._backward(loss, metrics)
        self._clip_grad_norm()
        self._step()
        return loss, metrics

    def predict_step(self, data):
        _, loss, metrics = self.model_fn(data)
        if self.num_gpu > 1:
            loss = loss.mean()
            for metric in metrics:
                metrics[metric] = metrics[metric].mean()
        return loss, metrics

    def _set_training_epoch_start(self, data_loader):
        if isinstance(data_loader, DataLoader) and isinstance(data_loader.sampler, DistributedSampler):
            data_loader.sampler.set_epoch(self.current_epoch)

        self.model_state = enums.ModelState.TRAIN
        self.train_state = enums.TrainingState.TRAIN_EPOCH_START
        self.model.train()
        if self.config.gradient_accumulation_steps > 1:
            self.optimizer.zero_grad()

    def _update_loss_metrics(self, losses, loss, metrics, data_loader):
        if self.model_state == enums.ModelState.TRAIN:
            losses.update(loss.item() * self.config.gradient_accumulation_steps, data_loader.batch_size)
        else:
            losses.update(loss.item(), data_loader.batch_size)

        if self.batch_index == 0:
            self.metrics_meter = {k: AverageMeter() for k in metrics}

        monitor = {}
        for m_m in self.metrics_meter:
            self.metrics_meter[m_m].update(metrics[m_m].cpu().detach().numpy(), data_loader.batch_size)
            monitor[m_m] = self.metrics_meter[m_m].avg
        if self.model_state == enums.ModelState.TRAIN:
            self.current_train_step += 1
        else:
            self.current_valid_step += 1
        self.update_metrics(losses=losses, monitor=monitor)
        return losses, monitor

    def _set_training_epoch_end(self, losses, monitor):
        self.update_metrics(losses=losses, monitor=monitor)
        self.train_state = enums.TrainingState.TRAIN_EPOCH_END

    def _set_training_state(self):
        self.model_state = enums.ModelState.TRAIN
        self.train_state = enums.TrainingState.TRAIN_EPOCH_START
        self.model.train()

    def train(self, data_loader):
        self._set_training_epoch_start(data_loader)
        losses = AverageMeter()
        for batch_index, data in enumerate(data_loader):
            self.batch_index = batch_index
            self.train_state = enums.TrainingState.TRAIN_STEP_START
            loss, metrics = self.train_step(data)
            losses, monitor = self._update_loss_metrics(losses, loss, metrics, data_loader)
            self.train_state = enums.TrainingState.TRAIN_STEP_END
            if self.valid_loader and self.config.val_strategy == "batch":
                if (
                    self.current_train_step % self.config.val_steps == 0
                    or self.current_train_step == self.num_train_steps
                ):
                    self.validate(self.valid_loader)
            if self._model_state.value == "end":
                break
        self._set_training_epoch_end(losses, monitor)

    def _set_validation_epoch_start(self, data_loader):
        if isinstance(data_loader, DataLoader) and isinstance(data_loader.sampler, DistributedSampler):
            data_loader.sampler.set_epoch(self.current_epoch)

        self.train_state = enums.TrainingState.VALID_EPOCH_START
        self.model_state = enums.ModelState.VALID
        self.model.eval()

    def _set_validation_epoch_end(self, losses, monitor):
        self.update_metrics(losses=losses, monitor=monitor)
        self.train_state = enums.TrainingState.VALID_EPOCH_END

    def validate(self, data_loader):
        self._set_validation_epoch_start(data_loader)
        losses = AverageMeter()

        for batch_index, data in enumerate(data_loader):
            self.batch_index = batch_index
            self.train_state = enums.TrainingState.VALID_STEP_START
            with torch.no_grad():
                loss, metrics = self.predict_step(data)
            losses, monitor = self._update_loss_metrics(losses, loss, metrics, data_loader)
            self.train_state = enums.TrainingState.VALID_STEP_END
        self._set_validation_epoch_end(losses, monitor)
        if self.config.val_strategy == "batch" and self._model_state.value != "end":
            self._set_training_state()

    def _step_scheduler_after_epoch(self):
        if self.scheduler is not None:
            if self.config.step_scheduler_after == "epoch":
                if self.config.step_scheduler_metric is None:
                    self.scheduler.step()
                else:
                    step_metric = self.name_to_metric(self.config.step_scheduler_metric)
                    self.scheduler.step(step_metric)

    def fit(self, train_dataset, valid_dataset=None, config: TezConfig = None, **kwargs):
        if config is None:
            config = TezConfig()
        self._init_trainer(train_dataset, valid_dataset, config, **kwargs)

        for _ in range(self.config.epochs):
            self.train_state = enums.TrainingState.EPOCH_START
            self.train(self.train_loader)
            if self.valid_loader and self.config.val_strategy == "epoch":
                self.validate(self.valid_loader)
            self._step_scheduler_after_epoch()
            self.train_state = enums.TrainingState.EPOCH_END
            if self._model_state.value == "end":
                time.sleep(2)
                break
            self.current_epoch += 1
        self.train_state = enums.TrainingState.TRAIN_END
        # TODO: do we need this?
        # if self.local_rank != -1:
        #    torch.distributed.barrier()

    def process_output(self, output):
        output = output.cpu().detach().numpy()
        return output

    def predict(self, dataset, **kwargs):

        if "sampler" in kwargs:
            sampler = kwargs["sampler"]
        else:
            sampler = None

        if "collate_fn" in kwargs:
            collate_fn = kwargs["collate_fn"]
        else:
            collate_fn = None

        if "batch_size" in kwargs:
            batch_size = kwargs["batch_size"]
        else:
            batch_size = self.config.test_batch_size

        if "num_jobs" in kwargs:
            num_jobs = kwargs["num_jobs"]
        else:
            num_jobs = self.config.num_jobs

        if "pin_memory" in kwargs:
            pin_memory = kwargs["pin_memory"]
        else:
            pin_memory = self.config.pin_memory

        if num_jobs == -1:
            num_jobs = multiprocessing.cpu_count()
            if num_jobs > 4:
                num_jobs -= 2

        if batch_size == 1:
            num_jobs = 0

        data_loader = DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=num_jobs,
            sampler=sampler,
            collate_fn=collate_fn,
            pin_memory=pin_memory,
        )

        if self.model.training:
            self.model.eval()

        for data in data_loader:
            with torch.no_grad():
                out, _, _ = self.model_fn(data)
                out = self.process_output(out)
                yield out



"""
Generic dataset for generic stuff
"""

import torch


class GenericDataset:
    def __init__(self, data, targets, dtypes):
        self.data = data
        self.targets = targets
        self.dtypes = dtypes

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        data = self.data[idx]
        targets = self.targets[idx]
        return {
            "x": torch.tensor(data, dtype=self.dtypes[0]),
            "y": torch.tensor(targets, dtype=self.dtypes[1]),
        }



"""
The tez model class
"""

import warnings

import psutil
import torch
import torch.nn as nn
from tqdm import tqdm

from tez import enums
from tez.callbacks import CallbackRunner
from tez.utils import AverageMeter


try:
    import torch_xla
    import torch_xla.core.xla_model as xm
    import torch_xla.distributed.xla_multiprocessing as xmp

    XLA_AVAILABLE = True
except ImportError:
    XLA_AVAILABLE = False


class Model(nn.Module):
    def __init__(self, *args, **kwargs):
        """
        Instead of inheriting from nn.Module, you import tez and inherit from tez.Model
        """
        super().__init__(*args, **kwargs)
        self.train_loader = None
        self.valid_loader = None
        self.optimizer = None
        self.scheduler = None
        self.step_scheduler_after = None
        self.step_scheduler_metric = None
        self.current_epoch = 0
        self.current_train_step = 0
        self.current_valid_step = 0
        self._model_state = None
        self._train_state = None
        self.device = None
        self._callback_runner = None
        self.fp16 = False
        self.scaler = None
        self.accumulation_steps = 0
        self.batch_index = 0
        self.metrics = {}
        self.metrics["train"] = {}
        self.metrics["valid"] = {}
        self.metrics["test"] = {}
        self.clip_grad_norm = None
        self.using_tpu = False

    @property
    def model_state(self):
        return self._model_state

    @model_state.setter
    def model_state(self, value):
        self._model_state = value
        # run something here in future if needed

    @property
    def train_state(self):
        return self._train_state

    @train_state.setter
    def train_state(self, value):
        self._train_state = value
        if self._callback_runner is not None:
            self._callback_runner(value)

    def name_to_metric(self, metric_name):
        if metric_name == "current_epoch":
            return self.current_epoch
        v_1 = metric_name.split("_")[0]
        v_2 = "_".join(metric_name.split("_")[1:])
        return self.metrics[v_1][v_2]

    def _init_model(
        self,
        device,
        train_dataset,
        valid_dataset,
        train_sampler,
        valid_sampler,
        train_bs,
        valid_bs,
        n_jobs,
        callbacks,
        fp16,
        train_collate_fn,
        valid_collate_fn,
        train_shuffle,
        valid_shuffle,
        accumulation_steps,
        clip_grad_norm,
    ):

        if callbacks is None:
            callbacks = list()

        if n_jobs == -1:
            n_jobs = psutil.cpu_count()

        self.device = device
        self.accumulation_steps = accumulation_steps
        self.clip_grad_norm = clip_grad_norm

        if next(self.parameters()).device != self.device:
            self.to(self.device)

        if self.train_loader is None:
            self.train_loader = torch.utils.data.DataLoader(
                train_dataset,
                batch_size=train_bs,
                num_workers=n_jobs,
                sampler=train_sampler,
                shuffle=train_shuffle,
                collate_fn=train_collate_fn,
            )
        if self.valid_loader is None:
            if valid_dataset is not None:
                self.valid_loader = torch.utils.data.DataLoader(
                    valid_dataset,
                    batch_size=valid_bs,
                    num_workers=n_jobs,
                    sampler=valid_sampler,
                    shuffle=valid_shuffle,
                    collate_fn=valid_collate_fn,
                )

        if self.optimizer is None:
            self.optimizer = self.fetch_optimizer()

        if self.scheduler is None:
            self.scheduler = self.fetch_scheduler()

        self.fp16 = fp16
        if self.fp16:
            self.scaler = torch.cuda.amp.GradScaler()

        self._callback_runner = CallbackRunner(callbacks, self)
        self.train_state = enums.TrainingState.TRAIN_START

    def monitor_metrics(self, *args, **kwargs):
        return

    def loss(self, *args, **kwargs):
        return

    def fetch_optimizer(self, *args, **kwargs):
        return

    def fetch_scheduler(self, *args, **kwargs):
        return

    def forward(self, *args, **kwargs):
        return super().forward(*args, **kwargs)

    def model_fn(self, data):
        for key, value in data.items():
            data[key] = value.to(self.device)
        if self.fp16:
            with torch.cuda.amp.autocast():
                output, loss, metrics = self(**data)
        else:
            output, loss, metrics = self(**data)
        return output, loss, metrics

    def train_one_step(self, data):
        if self.accumulation_steps == 1 and self.batch_index == 0:
            self.zero_grad()
        _, loss, metrics = self.model_fn(data)
        loss = loss / self.accumulation_steps
        if self.fp16:
            self.scaler.scale(loss).backward()
        else:
            loss.backward()
        if self.clip_grad_norm is not None:
            torch.nn.utils.clip_grad_norm_(self.parameters(), self.clip_grad_norm)
        if (self.batch_index + 1) % self.accumulation_steps == 0:
            if self.fp16:
                self.scaler.step(self.optimizer)
                self.scaler.update()
            else:
                if self.using_tpu:
                    xm.optimizer_step(self.optimizer, barrier=True)
                else:
                    self.optimizer.step()
            if self.scheduler:
                if self.step_scheduler_after == "batch":
                    if self.step_scheduler_metric is None:
                        self.scheduler.step()
                    else:
                        step_metric = self.name_to_metric(self.step_scheduler_metric)
                        self.scheduler.step(step_metric)
            if self.batch_index > 0:
                self.zero_grad()
        return loss, metrics

    def validate_one_step(self, data):
        _, loss, metrics = self.model_fn(data)
        return loss, metrics

    def predict_one_step(self, data):
        output, _, _ = self.model_fn(data)
        return output

    def update_metrics(self, losses, monitor):
        self.metrics[self._model_state.value].update(monitor)
        self.metrics[self._model_state.value]["loss"] = losses.avg

    def train_one_epoch(self, data_loader):
        self.train()
        self.model_state = enums.ModelState.TRAIN
        losses = AverageMeter()
        if self.accumulation_steps > 1:
            self.optimizer.zero_grad()
        if self.using_tpu:
            tk0 = data_loader
        else:
            tk0 = tqdm(data_loader, total=len(data_loader))
        for b_idx, data in enumerate(tk0):
            self.batch_index = b_idx
            self.train_state = enums.TrainingState.TRAIN_STEP_START
            loss, metrics = self.train_one_step(data)
            self.train_state = enums.TrainingState.TRAIN_STEP_END
            losses.update(loss.item() * self.accumulation_steps, data_loader.batch_size)
            if b_idx == 0:
                metrics_meter = {k: AverageMeter() for k in metrics}
            monitor = {}
            for m_m in metrics_meter:
                metrics_meter[m_m].update(metrics[m_m], data_loader.batch_size)
                monitor[m_m] = metrics_meter[m_m].avg
            self.current_train_step += 1
            if not self.using_tpu:
                tk0.set_postfix(loss=losses.avg, stage="train", **monitor)
            if self.using_tpu:
                print(f"train step: {self.current_train_step} loss: {losses.avg}")
        if not self.using_tpu:
            tk0.close()
        self.update_metrics(losses=losses, monitor=monitor)

        return losses.avg

    def validate_one_epoch(self, data_loader):
        self.eval()
        self.model_state = enums.ModelState.VALID
        losses = AverageMeter()
        if self.using_tpu:
            tk0 = data_loader
        else:
            tk0 = tqdm(data_loader, total=len(data_loader))
        for b_idx, data in enumerate(tk0):
            self.train_state = enums.TrainingState.VALID_STEP_START
            with torch.no_grad():
                loss, metrics = self.validate_one_step(data)
            self.train_state = enums.TrainingState.VALID_STEP_END
            losses.update(loss.item(), data_loader.batch_size)
            if b_idx == 0:
                metrics_meter = {k: AverageMeter() for k in metrics}
            monitor = {}
            for m_m in metrics_meter:
                metrics_meter[m_m].update(metrics[m_m], data_loader.batch_size)
                monitor[m_m] = metrics_meter[m_m].avg
            if not self.using_tpu:
                tk0.set_postfix(loss=losses.avg, stage="valid", **monitor)
            self.current_valid_step += 1
        if not self.using_tpu:
            tk0.close()
        self.update_metrics(losses=losses, monitor=monitor)
        return losses.avg

    def process_output(self, output):
        output = output.cpu().detach().numpy()
        return output

    def predict(self, dataset, sampler=None, batch_size=16, n_jobs=1, collate_fn=None):
        if next(self.parameters()).device != self.device:
            self.to(self.device)

        if n_jobs == -1:
            n_jobs = psutil.cpu_count()

        if batch_size == 1:
            n_jobs = 0
        data_loader = torch.utils.data.DataLoader(
            dataset, batch_size=batch_size, num_workers=n_jobs, sampler=sampler, collate_fn=collate_fn, pin_memory=True
        )

        if self.training:
            self.eval()

        if self.using_tpu:
            tk0 = data_loader
        else:
            tk0 = tqdm(data_loader, total=len(data_loader))

        for _, data in enumerate(tk0):
            with torch.no_grad():
                out = self.predict_one_step(data)
                out = self.process_output(out)
                yield out

            if not self.using_tpu:
                tk0.set_postfix(stage="test")

        if not self.using_tpu:
            tk0.close()

    def save(self, model_path, weights_only=False):
        model_state_dict = self.state_dict()
        if weights_only:
            if self.using_tpu:
                xm.save(model_state_dict, model_path)
            else:
                torch.save(model_state_dict, model_path)
            return
        if self.optimizer is not None:
            opt_state_dict = self.optimizer.state_dict()
        else:
            opt_state_dict = None
        if self.scheduler is not None:
            sch_state_dict = self.scheduler.state_dict()
        else:
            sch_state_dict = None
        model_dict = {}
        model_dict["state_dict"] = model_state_dict
        model_dict["optimizer"] = opt_state_dict
        model_dict["scheduler"] = sch_state_dict
        model_dict["epoch"] = self.current_epoch
        model_dict["fp16"] = self.fp16
        if self.using_tpu:
            xm.save(model_dict, model_path)
        else:
            torch.save(model_dict, model_path)

    def load(self, model_path, weights_only=False, device="cuda"):
        if device == "tpu":
            if XLA_AVAILABLE is False:
                raise RuntimeError("XLA is not available")
            else:
                self.using_tpu = True
                device = xm.xla_device()
        self.device = device
        if next(self.parameters()).device != self.device:
            self.to(self.device)
        model_dict = torch.load(model_path, map_location=torch.device(device))
        if weights_only:
            self.load_state_dict(model_dict)
        else:
            self.load_state_dict(model_dict["state_dict"])

    def fit(
        self,
        train_dataset,
        valid_dataset=None,
        train_sampler=None,
        valid_sampler=None,
        device="cuda",
        epochs=10,
        train_bs=16,
        valid_bs=16,
        n_jobs=8,
        callbacks=None,
        fp16=False,
        train_collate_fn=None,
        valid_collate_fn=None,
        train_shuffle=True,
        valid_shuffle=False,
        accumulation_steps=1,
        clip_grad_norm=None,
    ):
        print(
            "NOTE: This is old Model class and is deprecated. It will no longer be maintained! Please use version > 0.5.1. Its much better and supports multi-gpu training too!"
        )
        """
        The model fit function. Heavily inspired by tf/keras, this function is the core of Tez and this is the only
        function you need to train your models.

        """
        if device == "tpu":
            if XLA_AVAILABLE is False:
                raise RuntimeError("XLA is not available. Please install pytorch_xla")
            else:
                self.using_tpu = True
                fp16 = False
                device = xm.xla_device()
        self._init_model(
            device=device,
            train_dataset=train_dataset,
            valid_dataset=valid_dataset,
            train_sampler=train_sampler,
            valid_sampler=valid_sampler,
            train_bs=train_bs,
            valid_bs=valid_bs,
            n_jobs=n_jobs,
            callbacks=callbacks,
            fp16=fp16,
            train_collate_fn=train_collate_fn,
            valid_collate_fn=valid_collate_fn,
            train_shuffle=train_shuffle,
            valid_shuffle=valid_shuffle,
            accumulation_steps=accumulation_steps,
            clip_grad_norm=clip_grad_norm,
        )

        for _ in range(epochs):
            self.train_state = enums.TrainingState.EPOCH_START
            self.train_state = enums.TrainingState.TRAIN_EPOCH_START
            train_loss = self.train_one_epoch(self.train_loader)
            self.train_state = enums.TrainingState.TRAIN_EPOCH_END
            if self.valid_loader:
                self.train_state = enums.TrainingState.VALID_EPOCH_START
                valid_loss = self.validate_one_epoch(self.valid_loader)
                self.train_state = enums.TrainingState.VALID_EPOCH_END
            if self.scheduler:
                if self.step_scheduler_after == "epoch":
                    if self.step_scheduler_metric is None:
                        self.scheduler.step()
                    else:
                        step_metric = self.name_to_metric(self.step_scheduler_metric)
                        self.scheduler.step(step_metric)
            self.train_state = enums.TrainingState.EPOCH_END
            if self._model_state.value == "end":
                break
            self.current_epoch += 1
        self.train_state = enums.TrainingState.TRAIN_END



from enum import Enum


class ModelState(Enum):
    TRAIN = "train"
    VALID = "valid"
    TEST = "test"
    END = "end"


class TrainingState(Enum):
    TRAIN_START = "on_train_start"
    TRAIN_END = "on_train_end"
    EPOCH_START = "on_epoch_start"
    EPOCH_END = "on_epoch_end"
    TRAIN_EPOCH_START = "on_train_epoch_start"
    TRAIN_EPOCH_END = "on_train_epoch_end"
    VALID_EPOCH_START = "on_valid_epoch_start"
    VALID_EPOCH_END = "on_valid_epoch_end"
    TRAIN_STEP_START = "on_train_step_start"
    TRAIN_STEP_END = "on_train_step_end"
    VALID_STEP_START = "on_valid_step_start"
    VALID_STEP_END = "on_valid_step_end"
    TEST_STEP_START = "on_test_step_start"
    TEST_STEP_END = "on_test_step_end"



import cv2
import numpy as np
import torch
from PIL import Image, ImageFile


ImageFile.LOAD_TRUNCATED_IMAGES = True


class ImageDataset:
    def __init__(
        self,
        image_paths,
        targets,
        augmentations=None,
        backend="pil",
        channel_first=True,
        grayscale=False,
    ):
        """
        :param image_paths: list of paths to images
        :param targets: numpy array
        :param augmentations: albumentations augmentations
        """
        self.image_paths = image_paths
        self.targets = targets
        self.augmentations = augmentations
        self.backend = backend
        self.channel_first = channel_first
        self.grayscale = grayscale

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, item):
        targets = self.targets[item]
        if self.backend == "pil":
            image = Image.open(self.image_paths[item])
            image = np.array(image)
            if self.augmentations is not None:
                augmented = self.augmentations(image=image)
                image = augmented["image"]
        elif self.backend == "cv2":
            if self.grayscale is False:
                image = cv2.imread(self.image_paths[item])
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            else:
                image = cv2.imread(self.image_paths[item], cv2.IMREAD_GRAYSCALE)
            if self.augmentations is not None:
                augmented = self.augmentations(image=image)
                image = augmented["image"]
        else:
            raise Exception("Backend not implemented")
        if self.channel_first is True and self.grayscale is False:
            image = np.transpose(image, (2, 0, 1)).astype(np.float32)

        image_tensor = torch.tensor(image)
        if self.grayscale:
            image_tensor = image_tensor.unsqueeze(0)
        return {
            "image": image_tensor,
            "targets": torch.tensor(targets),
        }



import cv2
import numpy as np
import torch
from PIL import ImageFile


ImageFile.LOAD_TRUNCATED_IMAGES = True


class RCNNDataset:
    def __init__(
        self,
        image_paths,
        bounding_boxes,
        classes=None,
        augmentations=None,
        torchvision_format=True,
    ):
        self.image_paths = image_paths
        self.bounding_boxes = bounding_boxes
        self.augmentations = augmentations
        self.torchvision_format = torchvision_format
        self.classes = classes

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, item):
        # image = Image.open(self.image_paths[item])
        image = cv2.imread(self.image_paths[item])
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        bboxes = self.bounding_boxes[item]
        image = np.array(image)
        if self.augmentations is not None:
            augmented = self.augmentations(image=image, bboxes=bboxes)
            image = augmented["image"]
            bboxes = augmented["bboxes"]

        image = np.transpose(image, (2, 0, 1)).astype(np.float32)
        bboxes = np.array(bboxes)

        area = (bboxes[:, 3] - bboxes[:, 1]) * (bboxes[:, 2] - bboxes[:, 0])

        is_crowd = torch.zeros((bboxes.shape[0],), dtype=torch.int64)
        if self.classes is None:
            labels = torch.ones((bboxes.shape[0],), dtype=torch.int64)
        else:
            labels = torch.tensor(self.classes[item], dtype=torch.int64)

        target = {
            "boxes": torch.as_tensor(bboxes.tolist(), dtype=torch.float32),
            "area": torch.as_tensor(area.tolist(), dtype=torch.float32),
            "iscrowd": is_crowd,
            "labels": labels,
        }

        if self.torchvision_format:
            return torch.tensor(image, dtype=torch.float), target

        target["image"] = torch.tensor(image, dtype=torch.float)
        return target



from dataclasses import dataclass
from typing import Optional


@dataclass
class TezConfig:
    experiment_name = "default"
    device: Optional[str] = "cuda"  # cuda or cpu

    # batch sizes
    training_batch_size: Optional[int] = 32
    validation_batch_size: Optional[int] = 32
    test_batch_size: Optional[int] = 32

    # training parameters
    epochs: Optional[int] = 20
    gradient_accumulation_steps: Optional[int] = 1
    clip_grad_norm: Optional[float] = -1
    num_jobs: Optional[int] = -1
    fp16: Optional[bool] = False

    # data loader parameters
    train_shuffle: Optional[bool] = True
    valid_shuffle: Optional[bool] = True
    train_drop_last: Optional[bool] = False
    valid_drop_last: Optional[bool] = False
    test_drop_last: Optional[bool] = False
    test_shuffle: Optional[bool] = False
    pin_memory: Optional[bool] = True

    # scheduler parameters
    step_scheduler_after: Optional[str] = "epoch"  # "epoch" or "batch"
    step_scheduler_metric: Optional[str] = None

    # TODO: validation parameters
    val_strategy: Optional[str] = "epoch"  # epoch or batch
    val_steps: Optional[int] = 100  # not used if val_strategy is "epoch"



import numpy as np

from tez import enums
from tez.callbacks import Callback
from tez.logger import logger


class EarlyStopping(Callback):
    def __init__(self, monitor, model_path, patience=5, mode="min", delta=0.001, save_weights_only=False):
        self.monitor = monitor
        self.patience = patience
        self.counter = 0
        self.mode = mode
        self.best_score = None
        self.early_stop = False
        self.delta = delta
        self.save_weights_only = save_weights_only
        self.model_path = model_path
        if self.mode == "min":
            self.val_score = np.Inf
        else:
            self.val_score = -np.Inf

        if self.monitor.startswith("train_"):
            self.model_state = "train"
            self.monitor_value = self.monitor[len("train_") :]
        elif self.monitor.startswith("valid_"):
            self.model_state = "valid"
            self.monitor_value = self.monitor[len("valid_") :]
        else:
            raise Exception("monitor must start with train_ or valid_")

    def check(self, tez_trainer):
        epoch_score = tez_trainer.metrics[self.model_state][self.monitor_value]
        if self.mode == "min":
            score = -1.0 * epoch_score
        else:
            score = np.copy(epoch_score)

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(epoch_score, tez_trainer)
        elif score < self.best_score + self.delta:
            self.counter += 1
            logger.info("EarlyStopping counter: {} out of {}".format(self.counter, self.patience))
            if self.counter >= self.patience:
                tez_trainer.model_state = enums.ModelState.END
        else:
            self.best_score = score
            self.save_checkpoint(epoch_score, tez_trainer)
            self.counter = 0

    def on_valid_epoch_end(self, tez_trainer):
        if tez_trainer.config.val_strategy == "epoch":
            return
        self.check(tez_trainer)

    def on_epoch_end(self, tez_trainer):
        if tez_trainer.config.val_strategy == "batch":
            return
        self.check(tez_trainer)

    def save_checkpoint(self, epoch_score, tez_trainer):
        if epoch_score not in [-np.inf, np.inf, -np.nan, np.nan]:
            logger.info("\nScore improved ({} --> {}). Saving model!".format(self.val_score, epoch_score))
            tez_trainer.save(self.model_path, weights_only=self.save_weights_only)
        self.val_score = epoch_score



from setuptools import find_packages, setup


with open("README.md") as f:
    long_description = f.read()


if __name__ == "__main__":
    setup(
        name="tez",
        description="tez - a simple pytorch trainer",
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="Abhishek Thakur",
        author_email="abhishek4@gmail.com",
        url="https://github.com/abhishekkrthakur/tez",
        license="Apache License",
        packages=find_packages(),
        include_package_data=True,
        platforms=["linux", "unix"],
        python_requires=">=3.6",
    )



import os
import random

import numpy as np
import torch


def is_tpu_available():
    try:
        import torch_xla

        return True
    except ImportError:
        return False


class AverageMeter:
    """
    Computes and stores the average and current value
    """

    def __init__(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def seed_everything(seed: int):
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True
